#include "../../internal-include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.hpp>
#include <stdio.hpp>

namespace JabyEngine {
    namespace CDXA {
        CD::internal::State interrupt_handler(uint8_t irq);
    }

    namespace CD {
        namespace internal {
            struct File {
                uint32_t cur_lba;
                uint32_t dst_lba;

                void set_from(const AutoLBAEntry& file_info) {
                    this->cur_lba = file_info.get_lba();
                    this->dst_lba = this->cur_lba + file_info.get_size_in_sectors();
                }

                bool done_processing() {
                    this->cur_lba++;
                    return this->cur_lba == this->dst_lba;
                }
            };

            static constexpr auto AudioSectorMode   = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::AutoPauseTrack, CD_IO::Mode::CDDA);
            static constexpr auto DataSectorMode    = CD_IO::Mode::from(CD_IO::Mode::DoubleSpeed, CD_IO::Mode::DataSector);
            static constexpr auto XAAudioSectorMode = CD_IO::Mode::from(CD_IO::Mode::SingleSpeed, CD_IO::Mode::XADPCM, CD_IO::Mode::WholeSector, CD_IO::Mode::UseXAFilter);
            
            namespace IRQ {
                static SysCall::InterruptVerifierResult verifier();
                static uint32_t                         handler(uint32_t);
            }

            static SectorBufferAllocator sector_allocator;
            static File                  cur_file;
            
            volatile uint8_t          cmd_interrupt_bit = 0;
            State                     current_state     = State::Ready;
            SysCall::InterrupCallback callback          = {
                .next              = nullptr,
                .handler_function  = IRQ::handler,
                .verifier_function = IRQ::verifier
            };

            static BCDTimeStamp send_read_cmd0(uint32_t lba, CD_IO::Command::Desc desc) {
                const auto loc = BCDTimeStamp::from(lba);
                
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.min, loc.sec, loc.sector);
                Command::send<CD_IO::PortIndex0>(desc);
                return loc;
            }

            static void send_read_n0(uint32_t lba) {
                send_read_cmd0(lba, CD_IO::Command::ReadN);
                current_state = State::Reading;
            }

            namespace IRQ {
                static void read_sector_dma0(uint32_t* dst, size_t bytes) {
                    static const auto WaitSectorReady = []() {
                        while(!CD_IO::IndexStatus.read().is_set(CD_IO::IndexStatus::HasDataFifoData));
                    };

                    static const auto ReadSector = [](uint32_t* dst, size_t bytes) {
                        DMA_IO::CDROM.set_adr(reinterpret_cast<uintptr_t>(dst));
                        DMA_IO::CDROM.block_ctrl.write(DMA_IO::BCR::SyncMode0::for_cd(bytes >> 2));
                        DMA_IO::CDROM.channel_ctrl.write(DMA_IO::CHCHR::StartCDROM());
                        
                        DMA_IO::CDROM.wait();

                        CD_IO::PortIndex0::Request.write(CD_IO::Request::reset());
                    };

                    WaitSectorReady();
                    ReadSector(dst, bytes);
                }

                void read_sector_to0(uint32_t* dst, size_t bytes) {
                    CD_IO::PortIndex0::Request.write(CD_IO::Request::want_data());

                    // We only support DMA rn
                    read_sector_dma0(dst, bytes);

                    // Do we ever want to support reading via IO Port?
                    // Doesn't seem to important when we can use DMA
                }

                void resume_at0(const BCDTimeStamp& cd_time) {
                    Command::send<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, cd_time.min, cd_time.sec, cd_time.sector);

                    CD_IO::PortIndex1::change_to();
                    while(CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag) != CD_IO::Interrupt::Acknowledge);
                    CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);
                    CD_IO::PortIndex0::change_to();
                }

                //######################################################################################################################

                static SysCall::InterruptVerifierResult verifier() {
                    if(Interrupt::is_irq(Interrupt::CDROM)) {
                        return SysCall::InterruptVerifierResult::ExecuteHandler;
                    }
                    
                    else {
                        return SysCall::InterruptVerifierResult::SkipHandler;
                    }
                }

                static uint32_t handler(uint32_t) {
                    const auto old_status = CD_IO::IndexStatus.read();

                    CD_IO::PortIndex1::change_to();
                    const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                    CD_IO::Interrupt::ack_extended(CD_IO::PortIndex1::InterruptFlag);
                    CD_IO::PortIndex0::change_to();

                    cmd_interrupt_bit = bit::clear(cmd_interrupt_bit, cur_irq);

                    if(current_state != State::XAMode) {
                        switch(cur_irq) {
                            case CD_IO::Interrupt::DataReady: {
                                // Obtain sector content here
                                auto* sector = sector_allocator.allocate_sector();
                                if(sector) {
                                    //Now obtain sector
                                    read_sector_to0(sector->data, CD_IO::DataSector::SizeBytes);

                                    if(cur_file.done_processing()) {
                                        current_state = State::Done;
                                        pause();
                                    }
                                }

                                else {
                                    current_state = State::BufferFull;
                                    pause();
                                }
                            } break;

                            case CD_IO::Interrupt::DataEnd: {
                                // TODO: Fix this!! This is a freaking static time
                                resume_at0(BCDTimeStamp{.min = 0x0, .sec = 0x09, .sector = 0x0});
                                Command::send<CD_IO::PortIndex0>(CD_IO::Command::Play);
                            } break;

                            case CD_IO::Interrupt::DiskError: {
                                current_state = State::Error;
                            } break;
                        }
                    }

                    else {
                        current_state = CDXA::interrupt_handler(cur_irq);
                    }

                    // No masking required because we can only write bit 0 - 2
                    CD_IO::IndexStatus.write(old_status);
                    Interrupt::ack_irq(Interrupt::CDROM);

                    SysCall::ReturnFromException();
                    __builtin_unreachable();
                }
            }

            void read_file(AutoLBAEntry file_info, const SectorBufferAllocator& buffer_allocator) {
                cur_file.set_from(file_info);
                sector_allocator = buffer_allocator;

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);
                send_read_n0(cur_file.cur_lba);
            }


            void continue_reading() {
                if(current_state == State::BufferFull) {
                    Command::wait_completed();
                    CD_IO::PortIndex0::change_to();
                    send_read_n0(cur_file.cur_lba);    
                }                
            }

            BCDTimeStamp get_loc() {
                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::GetLocP);

                const auto track   = CD_IO::PortIndex0::ResponseFifo.read().raw; // track number (AAh=Lead-out area) (FFh=unknown, toc, none?)
                const auto index   = CD_IO::PortIndex0::ResponseFifo.read().raw; // index number (Usually 01h)
                const auto mm      = CD_IO::PortIndex0::ResponseFifo.read().raw; // minute number within track (00h and up)
                const auto ss      = CD_IO::PortIndex0::ResponseFifo.read().raw; // second number within track (00h to 59h)
                const auto sect    = CD_IO::PortIndex0::ResponseFifo.read().raw; // sector number within track (00h to 74h)

                const auto min     = CD_IO::PortIndex0::ResponseFifo.read().raw; // minute number on entire disk (00h and up)
                const auto sec     = CD_IO::PortIndex0::ResponseFifo.read().raw; // second number on entire disk (00h to 59h)
                const auto sectors = CD_IO::PortIndex0::ResponseFifo.read().raw; // sector number on entire disk (00h to 74h) 
                return BCDTimeStamp{min, sec, sectors};
            }

            void enable_CDDA() {
                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, AudioSectorMode);
            }

            void enable_CDXA(bool double_speed) {
                static constexpr uint8_t SingleSpeedBit = 0x0;
                static constexpr uint8_t DoubleSpeedBit = static_cast<uint8_t>(CD_IO::Mode::DoubleSpeed);

                const uint8_t mode = XAAudioSectorMode.raw | (double_speed ? DoubleSpeedBit : SingleSpeedBit);

                Command::wait_completed();
                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, mode);
            }
        }
    }
}